@@ -17,6 +17,7 @@
 
 #include "atom.h"
 #include "eat_file.h"
+#include "hash.h"
 #include "rmspace.h"
 #include "scandirat.h"
 #include "set.h"
@@ -27,7 +28,7 @@
 #include <xalloc.h>
 
 static int tree_pkg_compar(const void *l, const void *r);
-static tree_pkg_ctx * tree_next_pkg_int(tree_cat_ctx *cat_ctx);
+static tree_pkg_ctx *tree_next_pkg_int(tree_cat_ctx *cat_ctx);
 static void tree_close_meta(tree_pkg_meta *cache);
 
 static tree_ctx *
@@ -132,18 +133,19 @@ tree_ctx *
 tree_open_binpkg(const char *sroot, const char *spkg)
 {
 	tree_ctx *ret = tree_open_int(sroot, spkg, true);
-	char buf[_Q_PATH_MAX];
+	int fd;
 
 	if (ret != NULL) {
 		ret->cachetype = CACHE_BINPKGS;
 
-		snprintf(buf, sizeof(buf), "%s%s/%s", sroot, spkg, binpkg_packages);
-		if (eat_file(buf, &ret->pkgs, &ret->pkgslen)) {
+		fd = openat(ret->tree_fd, binpkg_packages, O_RDONLY | O_CLOEXEC);
+		if (eat_file_fd(fd, &ret->pkgs, &ret->pkgslen)) {
 			ret->cachetype = CACHE_PACKAGES;
 		} else if (ret->pkgs != NULL) {
 			free(ret->pkgs);
 			ret->pkgs = NULL;
 		}
+		close(fd);
 	}
 
 	return ret;
@@ -386,20 +388,18 @@ tree_filter_pkg(const struct dirent *de)
 tree_pkg_ctx *
 tree_open_pkg(tree_cat_ctx *cat_ctx, const char *name)
 {
-	tree_pkg_ctx *pkg_ctx = xmalloc(sizeof(*pkg_ctx));
+	tree_pkg_ctx *pkg_ctx = xzalloc(sizeof(*pkg_ctx));
 	pkg_ctx->name = name;
-	pkg_ctx->slot = NULL;
 	pkg_ctx->repo = cat_ctx->ctx->repo;
 	pkg_ctx->fd = -1;
 	pkg_ctx->cat_ctx = cat_ctx;
-	pkg_ctx->atom = NULL;
-	pkg_ctx->meta = NULL;
 
 	/* see if this pkg matches the query, here we can finally check
 	 * version conditions like >=, etc. */
 	if (cat_ctx->ctx->query_atom != NULL) {
 		(void)tree_get_atom(pkg_ctx, cat_ctx->ctx->query_atom->SLOT != NULL);
 		if (atom_compare(pkg_ctx->atom, cat_ctx->ctx->query_atom) != EQUAL) {
+			pkg_ctx->name = NULL;
 			tree_close_pkg(pkg_ctx);
 			return NULL;
 		}
@@ -957,9 +957,8 @@ tree_read_file_binpkg_xpak_cb(
 
 	if (len - pos < (size_t)(data_len + 1)) {
 		char *old_data = m->Q__data;
-		len += (((data_len + 1) / BUFSIZ) + 1) * BUFSIZ;
+		len += (((data_len + 1 - (len - pos)) / BUFSIZ) + 1) * BUFSIZ;
 		m->Q__data = xrealloc(m->Q__data, len);
-		m->Q__md5_ = (char *)len;
 
 		/* re-position existing keys */
 		if (old_data != NULL && m->Q__data != old_data) {
@@ -969,6 +968,10 @@ tree_read_file_binpkg_xpak_cb(
 				if (newdata[elems] != NULL)
 					newdata[elems] = m->Q__data + (newdata[elems] - old_data);
 		}
+
+		/* set after repositioning! */
+		m->Q__md5_ = (char *)len;
+		m->Q__eclasses_ = (char *)pos;
 	}
 
 	*key = m->Q__data + pos;
@@ -981,10 +984,51 @@ static tree_pkg_meta *
 tree_read_file_binpkg(tree_pkg_ctx *pkg_ctx)
 {
 	tree_pkg_meta *m = xzalloc(sizeof(tree_pkg_meta));
+	int newfd = dup(pkg_ctx->fd);
 
 	xpak_process_fd(pkg_ctx->fd, true, m, tree_read_file_binpkg_xpak_cb);
 	pkg_ctx->fd = -1;  /* closed by xpak_process_fd */
 
+	/* fill in some properties which are not available, but would be in
+	 * Packages, and used to verify the package ... this is somewhat
+	 * fake, but allows to transparantly use a dir of binpkgs */
+	if (newfd != -1) {
+		size_t fsize;
+		size_t needlen = 40 + 1 + 19 + 1;
+		size_t pos = (size_t)m->Q__eclasses_;
+		size_t len = (size_t)m->Q__md5_;
+
+		if (len - pos < needlen) {
+			char *old_data = m->Q__data;
+			len += (((needlen - (len - pos)) / BUFSIZ) + 1) * BUFSIZ;
+			m->Q__data = xrealloc(m->Q__data, len);
+
+			/* re-position existing keys */
+			if (old_data != NULL && m->Q__data != old_data) {
+				char **newdata = (char **)m;
+				int elems = sizeof(tree_pkg_meta) / sizeof(char *);
+				while (elems-- > 1)  /* skip Q__data itself */
+					if (newdata[elems] != NULL)
+						newdata[elems] =
+							m->Q__data + (newdata[elems] - old_data);
+			}
+
+			/* set after repositioning! */
+			m->Q__md5_ = (char *)len;
+			m->Q__eclasses_ = (char *)pos;
+		}
+
+		m->Q_SHA1 = m->Q__data + pos;
+		m->Q_SIZE = m->Q_SHA1 + 40 + 1;
+		pos += needlen;
+		m->Q__eclasses_ = (char *)pos;
+
+		lseek(newfd, 0, SEEK_SET);  /* reposition at the whole file */
+		if (hash_multiple_file_fd(newfd, NULL, m->Q_SHA1, NULL, NULL,
+				NULL, NULL, &fsize, HASH_SHA1) == 0)
+			snprintf(m->Q_SIZE, 19 + 1, "%zu", fsize);
+	}
+
 	return m;
 }
 
@@ -1074,9 +1118,8 @@ tree_pkg_meta_get_int(tree_pkg_ctx *pkg_ctx, size_t offset, const char *keyn)
 			/* TODO: this is an exact copy from tree_read_file_binpkg_xpak_cb */
 			if (len - pos < (size_t)(s.st_size + 1)) {
 				p = m->Q__data;
-				len += (((s.st_size + 1) / BUFSIZ) + 1) * BUFSIZ;
+				len += (((s.st_size + 1 - (len - pos)) / BUFSIZ) + 1) * BUFSIZ;
 				m->Q__data = xrealloc(m->Q__data, len);
-				m->Q__md5_ = (char *)len;
 
 				/* re-position existing keys */
 				if (p != NULL && m->Q__data != p) {
@@ -1086,6 +1129,10 @@ tree_pkg_meta_get_int(tree_pkg_ctx *pkg_ctx, size_t offset, const char *keyn)
 						if (newdata[elems] != NULL)
 							newdata[elems] = m->Q__data + (newdata[elems] - p);
 				}
+
+				/* set after repositioning! */
+				m->Q__md5_ = (char *)len;
+				m->Q__eclasses_ = (char *)pos;
 			}
 
 			p = *key = m->Q__data + pos;
@@ -1239,11 +1286,11 @@ tree_close_pkg(tree_pkg_ctx *pkg_ctx)
 static int
 tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 {
-	char *p = ctx->pkgs;
+	char *p;
 	char *q;
 	char *c;
 	char pkgname[_Q_PATH_MAX];
-	size_t len = ctx->pkgslen;
+	size_t len;
 	int ret = 0;
 	depend_atom *query = ctx->query_atom;
 
@@ -1253,6 +1300,25 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 	tree_pkg_meta meta;
 	depend_atom *atom = NULL;
 
+	/* re-read the contents, this is necessary to make it possible to
+	 * call this function multiple times
+	 * TODO: generate an internal in-memory tree when cache is enabled */
+	if (ctx->pkgs == NULL || ctx->pkgs[0] == '\0') {
+		int fd = openat(ctx->tree_fd, binpkg_packages, O_RDONLY | O_CLOEXEC);
+		if (!eat_file_fd(fd, &ctx->pkgs, &ctx->pkgslen)) {
+			if (ctx->pkgs != NULL) {
+				free(ctx->pkgs);
+				ctx->pkgs = NULL;
+			}
+			close(fd);
+			return 1;
+		}
+		close(fd);
+	}
+
+	p = ctx->pkgs;
+	len = strlen(ctx->pkgs);  /* sucks, need eat_file change */
+
 	memset(&meta, 0, sizeof(meta));
 
 	do {
@@ -1349,12 +1415,14 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 		match_key(IUSE);
 		match_key(KEYWORDS);
 		match_key(LICENSE);
-		match_key2(MD5, _md5_);
-		match_key2(SHA1, _eclasses_);
+		match_key(MD5);
+		match_key(SHA1);
 		match_key(RDEPEND);
 		match_key(SLOT);
 		match_key(USE);
 		match_key(PDEPEND);
+		match_key2(REPO, repository);
+		match_key(SIZE);
 #undef match_key
 #undef match_key2
 		}
@@ -1374,6 +1442,7 @@ tree_foreach_packages(tree_ctx *ctx, tree_pkg_cb callback, void *priv)
 	/* ensure we don't free a garbage pointer */
 	ctx->repo = NULL;
 	ctx->do_sort = false;
+	ctx->pkgs[0] = '\0';
 
 	return ret;
 }
@@ -1408,6 +1477,8 @@ tree_foreach_pkg(tree_ctx *ctx, tree_pkg_cb callback, void *priv,
 	/* allow foreach to be called again on the same open tree */
 	if (ctx->do_sort)
 		scandir_free(ctx->cat_de, ctx->cat_cnt);
+	else
+		rewinddir(ctx->dir);
 	ctx->cat_de = NULL;
 	ctx->cat_cur = 0;
 	ctx->cat_cnt = 0;
@@ -1524,34 +1595,167 @@ tree_get_atoms(tree_ctx *ctx, bool fullcpv, set *satoms)
 	return state.cpf;
 }
 
-tree_pkg_ctx *
-tree_match_atom(tree_ctx *ctx, depend_atom *a)
+struct tree_match_pkgs_cb_ctx {
+	int flags;
+	tree_match_ctx *ret;
+};
+
+static int
+tree_match_atom_packages_cb(tree_pkg_ctx *ctx, void *priv)
+{
+	struct tree_match_pkgs_cb_ctx *rctx = priv;
+	depend_atom *a;
+	tree_match_ctx *n;
+
+	/* skip anything after finding first match */
+	if (rctx->flags & TREE_MATCH_FIRST && rctx->ret != NULL)
+		return 1;
+
+	a = tree_get_atom(ctx, rctx->flags & TREE_MATCH_FULL_ATOM);
+
+	/* skip virtual category if not requested */
+	if (!(rctx->flags & TREE_MATCH_VIRTUAL) &&
+			strcmp(a->CATEGORY, "virtual") == 0)
+		return 1;
+
+	n = xzalloc(sizeof(tree_match_ctx));
+	n->free_atom = true;
+	n->atom = atom_clone(a);
+	if (rctx->flags & TREE_MATCH_METADATA) {
+		n->meta = xmalloc(sizeof(*n->meta));
+		/* for Packages, all pointers to meta here are to the in memory
+		 * copy of the Packages file, so these pointers can just be
+		 * copied since the tree has to remain open, thus the pointers
+		 * will stay valid */
+		memcpy(n->meta, ctx->cat_ctx->ctx->pkgs, sizeof(*n->meta));
+	}
+
+	n->next = rctx->ret;
+	rctx->ret = n;
+
+	return 0;
+}
+
+static int
+tree_match_atom_binpkg_cb(tree_pkg_ctx *ctx, void *priv)
+{
+	struct tree_match_pkgs_cb_ctx *rctx = priv;
+	depend_atom *a;
+	tree_match_ctx *n;
+
+	/* skip anything after finding first match */
+	if (rctx->flags & TREE_MATCH_FIRST && rctx->ret != NULL)
+		return 1;
+
+	a = tree_get_atom(ctx, rctx->flags & TREE_MATCH_FULL_ATOM);
+
+	/* skip virtual category if not requested */
+	if (!(rctx->flags & TREE_MATCH_VIRTUAL) &&
+			strcmp(a->CATEGORY, "virtual") == 0)
+		return 1;
+
+	n = xzalloc(sizeof(tree_match_ctx));
+	n->free_atom = true;
+	n->atom = atom_clone(a);
+	if (rctx->flags & TREE_MATCH_METADATA)
+		n->meta = tree_pkg_read(ctx);
+
+	n->next = rctx->ret;
+	rctx->ret = n;
+
+	return 0;
+}
+
+tree_match_ctx *
+tree_match_atom(tree_ctx *ctx, depend_atom *query, int flags)
 {
 	tree_cat_ctx *cat_ctx;
 	tree_pkg_ctx *pkg_ctx;
+	tree_match_ctx *ret = NULL;
 	depend_atom *atom;
 
+	if (ctx->cachetype == CACHE_PACKAGES) {
+		struct tree_match_pkgs_cb_ctx rctx;
+		/* Packages needs to be serviced separately because it doesn't
+		 * use a tree internally, but reads off of the Packages file */
+		rctx.flags = flags;
+		rctx.ret = NULL;
+		ctx->query_atom = query;
+		tree_foreach_packages(ctx, tree_match_atom_packages_cb, &rctx);
+		ctx->query_atom = NULL;
+		return rctx.ret;
+	} else if (ctx->cachetype == CACHE_BINPKGS) {
+		struct tree_match_pkgs_cb_ctx rctx;
+		/* this sulks, but binpkgs modify the pkg_ctx->name to strip off
+		 * .tbz2, and that makes it non-reusable */
+		rctx.flags = flags;
+		rctx.ret = NULL;
+		tree_foreach_pkg(ctx, tree_match_atom_binpkg_cb, &rctx, true, query);
+		return rctx.ret;
+	}
+
+	/* activate cache for future lookups, tree_match_atom relies on
+	 * cache behaviour from tree, which means all categories and
+	 * packages remain in memory until tree_close is being called */
 	if (ctx->cache.categories == NULL)
 		ctx->cache.categories = create_set();
 
-	if (a->P == NULL) {
-		return NULL;
-	} else if (a->CATEGORY == NULL) {
-		/* loop through all cats and recurse */
-		/* TODO: some day */
-		return NULL;
+	ctx->do_sort = true;     /* sort uses buffer, which cache relies on */
+	ctx->query_atom = NULL;  /* ensure the cache contains ALL pkgs */
+
+#define search_cat(C) \
+{ \
+	while ((pkg_ctx = tree_next_pkg(C)) != NULL) { \
+		atom = tree_get_atom(pkg_ctx, \
+				query->SLOT != NULL || flags & TREE_MATCH_FULL_ATOM); \
+fprintf(stderr, "fbg: %s\n", atom_to_string(atom)); \
+		if (flags & TREE_MATCH_VIRTUAL || \
+				strcmp(atom->CATEGORY, "virtual") != 0) \
+			if (atom_compare(atom, query) == EQUAL) { \
+				tree_match_ctx *n; \
+				n = xzalloc(sizeof(tree_match_ctx)); \
+				n->free_atom = false; \
+				n->atom = atom; \
+				if (flags & TREE_MATCH_METADATA) \
+					n->meta = tree_pkg_read(pkg_ctx); \
+				n->next = ret; \
+				ret = n; \
+			} \
+			if (flags & TREE_MATCH_FIRST && ret != NULL) \
+				break; \
+	} \
+	C->pkg_cur = 0;  /* reset to allow another traversal */ \
+}
+
+	if (query->CATEGORY == NULL) {
+		/* loop through all cats */
+		while ((cat_ctx = tree_next_cat(ctx)) != NULL) {
+			search_cat(cat_ctx);
+			if (ret != NULL && flags & TREE_MATCH_FIRST)
+				break;
+		}
+		/* allow running again through the cats */
+		ctx->cat_cur = 0;
 	} else {
 		/* try CAT, and PN for latest version */
-		if ((cat_ctx = tree_open_cat(ctx, a->CATEGORY)) == NULL)
-			return NULL;
-		ctx->do_sort = true;     /* sort uses buffer, which cache relies on */
-		ctx->query_atom = NULL;  /* ensure the cache contains ALL pkgs */
-		while ((pkg_ctx = tree_next_pkg(cat_ctx)) != NULL) {
-			atom = tree_get_atom(pkg_ctx, a->SLOT != NULL);
-			if (atom_compare(atom, a) == EQUAL)
-				return pkg_ctx;
-		}
+		if ((cat_ctx = tree_open_cat(ctx, query->CATEGORY)) != NULL)
+			search_cat(cat_ctx);
+	}
 
-		return NULL;
+	return ret;
+}
+
+void
+tree_match_close(tree_match_ctx *match)
+{
+	tree_match_ctx *w;
+
+	for (w = NULL; match != NULL; match = w) {
+		w = match->next;
+		if (match->free_atom)
+			atom_implode(match->atom);
+		if (match->meta != NULL)
+			tree_close_meta(match->meta);
+		free(match);
 	}
 }
