@@ -594,15 +594,17 @@ env_vars vars_to_read[] = {
 #define _Q_EV(t, V, set, lset, d) \
 { \
 	.name = #V, \
-	.name_len = strlen(#V), \
+	.name_len = sizeof(#V) - 1, \
 	.type = _Q_##t, \
 	set, \
 	lset, \
 	.default_value = d, \
 	.src = NULL, \
 },
-#define _Q_EVS(t, V, v, d) _Q_EV(t, V, .value.s = &v, .value_len = strlen(d), d)
-#define _Q_EVB(t, V, v, d) _Q_EV(t, V, .value.b = &v, .value_len = 0, d)
+#define _Q_EVS(t, V, v, d) \
+	_Q_EV(t, V, .value.s = &v, .value_len = sizeof(d) - 1, d)
+#define _Q_EVB(t, V, v, d) \
+	_Q_EV(t, V, .value.b = &v, .value_len = 0, d)
 
 	_Q_EVS(STR,  ROOT,                portroot,            "/")
 	_Q_EVS(STR,  ACCEPT_LICENSE,      accept_license,      "")
@@ -624,6 +626,8 @@ env_vars vars_to_read[] = {
 	{ NULL, 0, _Q_BOOL, { NULL }, 0, NULL, NULL, }
 
 #undef _Q_EV
+#undef _Q_EVS
+#undef _Q_EVB
 };
 set *package_masks = NULL;
 
